/*==========================================================================
//  Dcryptage du chiffrement de Vigenre
//  Projet TER, Matrise d'Informatique 1997,
//  Universit de Bordeaux 1, France,
//  Tous droits rservs.
//
//  de Maillard Luc, Montiton Frdric
//  Morreeuw Jean-Franois et Lompo Issa
//
//  mailto:morreeuw@emi.u-bordeaux.fr,dbv@komicom.remcomp.fr
//  http://www.emi.u_bordeaux.fr/~morreeuw/universite/TER
//
//  MODULE DE DECRYPTAGE AUTOMATIQUE
//
//========================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include "auto.h"

static int      _max(int value1, int value2);
static int      _min(int value1, int value2);
static result_p _doubloon(char *data, long length, int depth, short ratio);
static void     _coincide(char *data, long length, result_p result);
static result_p _minter(char *data, long length, result_p passlen, int depth);
static result_p _offset(char *data, long length, result_p passlen,
			result_p passoff, int depth);

static result_p offset_table[AUTO_DEPTH_MAX];
static char *   password_table[AUTO_DEPTH_MAX];

/*==========================================================================
//  lecture des fichiers de donnees frequentielles des langues
//========================================================================*/

int
auto_install(char *path)
{
  int number = shift_install(path);

  if(number==0)
    RAISE(io_XCEPT, path, "No language file");
  return number;
}

/*==========================================================================
//  fonction de dechiffrement automatique
//========================================================================*/

char **
auto_decipher(char *data, long length, int depth, short ratio)
{
  int i;
  char **password = (char **) malloc2((2*depth+1)*sizeof(*password));
  result_p passoff = NULL;
  result_p passlst = NULL;
  result_p passlen = NULL;
  int autodepth = _min(AUTO_DEPTH_MAX, _max(AUTO_DEPTH_MIN, depth));

  for(i=0; i<2*depth+1; i++)
    password[i] = NULL;

  cipher_define(CHAR_MIN, CHAR_MAX);

  passlen = _doubloon(data, length, autodepth, ratio);
  _coincide(data, length, passlen);
  passoff = _minter(data, length, passlen, 2*autodepth);
  passlst = _offset(data, length, passlen, passoff, autodepth);

  for(i=0; i<_min(result_depth(passlst), depth); i++)
    {
      short *offset = result_offset(passlst, i);
      short width = offset[2];
      short language = offset[4];
      char *basepass = password_table[offset[0]];

      {
        char shift[2] = {'\0', '\0'};
        shift[0] = offset[3]+CHAR_MIN;
        password[2*i] = cipher_compute(shift, basepass, width+1, 1);
        password[2*i][width]='\0';
      }

      password[2*i+1] = (char *) strdup2(shift_language(language));
    }

  for(i=0; i<result_depth(passlen); i++)
    if(offset_table[i])
      result_destroy(offset_table[i]);

  for(i=0; i<_min(result_depth(passoff), autodepth); i++)
    free2(password_table[i]);

  result_destroy(passlen);
  result_destroy(passoff);
  result_destroy(passlst);

  return password;
}

/*========================================================================*/

int
_max(int value1, int value2)
{
  return (value1>value2)?value1:value2;
}

/*------------------------------------------------------------------------*/

int
_min(int value1, int value2)
{
  return (value1<value2)?value1:value2;
}

/*========================================================================*/

static double mapfun1(short *offset, double quality,
		      short width, void *argument);

/*------------------------------------------------------------------------*/

double
mapfun1(short *offset, double quality, short width, void *argument)
{
  if(offset[0] > *((short *) argument))
    return 0;
  else
    return quality;
}

/*------------------------------------------------------------------------*/

#define KH_DIVIDE_LIMIT .7

result_p
_doubloon(char *data, long length, int depth, short ratio)
{
  int buffer = 2048;
  int step = _max(2, _min(6, length/10));
  result_p result;
  short limit = _max(10, _min(AUTO_KH_LIMIT, length/ratio));

  while(result_depth(result = search_key_length(data, length,
						step, buffer, depth))<=3
        && step>2)
    {
      result_destroy(result);
      step--;
    }

  {
    int i;
    short best_offset = *(result_offset(result, 0));
    double best_quality = *(result_offset(result, 0));
    double limit_quality = ic_coincide(data, length, best_offset);
    double new_quality = 0;
    result_p tmp_result = result_create(depth, 1);

    for(i=0; i<result_depth(result); i++)
      {
        short *newoffset = result_newoffset(tmp_result);
        double *newquality = result_newquality(tmp_result);
        *newoffset = *(result_offset(result, i));
        *newquality = *(result_quality(result, i));
        result_insert(tmp_result);
      }

    result_destroy(result);
    result = tmp_result;

    for(i=best_offset-1; i>0; i--)
      if((best_offset%i)==0)
        if((new_quality = ic_coincide(data, length, i))
	                      >= limit_quality*KH_DIVIDE_LIMIT)
          {
            short *newoffset = result_newoffset(result);
            double *newquality = result_newquality(result);
            newoffset[0]=i;
            *newquality = best_quality;
            result_insert(result);
          }
  }

  result_map(result, mapfun1, (void *) &limit);
  result_sort(result);
  result_clean(result);
  if(result_depth(result)<2)
    {
      int i;
      result_destroy(result);
      result = result_create(depth, 1);
      for(i=0; i<_min(1.5*limit, length); i++)
        {
          short *newoffset = result_newoffset(result);
          double *newquality = result_newquality(result);
          newoffset[0]=i+1;
          *newquality = ic_coincide(data, length, i+1);
          result_insert(result);
        }
      result_sort(result);
      if(result_clean(result)==0)
        RAISE(fatal_XCEPT, "doubloon", "Cannot find any doubloon");
    }

  return result;
}

/*========================================================================*/

struct maparg
{
  char * data;
  long   length;
};

static double mapfun2(short *offset, double quality,
		      short width, void *argument);

/*------------------------------------------------------------------------*/

double
mapfun2(short *offset, double quality, short width, void *argument)
{
  struct maparg *arg = (struct maparg *) argument;
  double value = ic_coincide(arg->data, arg->length, offset[0]);
  return quality*value*value;
}

/*------------------------------------------------------------------------*/

void
_coincide(char *data, long length, result_p passlen)
{
  struct maparg argument;
  argument.data = data;
  argument.length = length;

  result_map(passlen, mapfun2, (void *) &argument);
  result_sort(passlen);
  result_clean(passlen);

  if(result_depth(passlen)==0)
    RAISE(fatal_XCEPT, NULL, NULL);

  return;
}

/*========================================================================*/

result_p
_minter(char *data, long length, result_p passlen, int depth)
{
  int i;
  result_p result = result_create(depth, 2);

  for(i=0; i<result_depth(passlen); i++)
    {
      int j;
      short stopit = 0;
      short *offset = result_offset(passlen, i);
      double quality = *result_quality(passlen, i);
      short cycle = offset[0];
      
      offset_table[i] = NULL;
      HANDLE(fatal_XCEPT,
	     offset_table[i] = kh_compute(data, length, cycle, depth),
	     stopit=-1);

      for(j=0; stopit!=-1 && j<result_depth(offset_table[i]); j++)
        {
          short *newoffset = result_newoffset(result);
          double *newquality = result_newquality(result);
          double tmpquality = *result_quality(offset_table[i], j);

          newoffset[0] = i;
          newoffset[1] = j;
          newquality[0] = quality*tmpquality;

          stopit = result_insert(result);
        }
    }

  result_clean(result);
  if(result_depth(result)==0)
    RAISE(fatal_XCEPT, NULL, NULL);
  return result;
}

/*========================================================================*/

result_p
_offset(char *data, long length, result_p passlen, result_p passoff, int depth)
{
  int i;
  result_p result = result_create(depth, 5);

  for(i=0; i<_min(result_depth(passoff), depth); i++)
    {
      int j;
      short *tmpoff = result_offset(passoff, i);
      short *tmplen = result_offset(passlen, tmpoff[0]);
      short width = tmplen[0];
      short *offset = result_offset(offset_table[tmpoff[0]], tmpoff[1]);
      char *password = (char *) malloc2((width+1)*sizeof(*password));

      password_table[i] = password;
      password[width]='\0';

      {
        short tmp = 0;
        password[0]=CHAR_MIN;
        for(j=1; j<width; j++)
          password[j] = (tmp = (tmp+offset[j-1])%CHAR_RANGE) +CHAR_MIN;
      }

      {
        short stopit = 0;
        double quality = *result_quality(passoff, i);
        char *newdata = cipher_compute(password, data, length, -1);
        result_p tmpresult = shift_search(newdata, length, depth);

        for(j=0; stopit!=-1 && j<result_depth(tmpresult); j++)
          {
            short *tmpoffset = result_offset(tmpresult, j);
            double tmpquality = *result_quality(tmpresult, j);

            short *newoffset = result_newoffset(result);
            double *newquality = result_newquality(result);

            newoffset[0] = i;                   /* numero du mot de base */
            newoffset[1] = j;                   /* position dans l'ordre */
            newoffset[2] = width;               /* longueur du password */
            newoffset[3] = tmpoffset[0];        /* decalage */
            newoffset[4] = tmpoffset[1];        /* langue utilisee */
            newquality[0] = quality*tmpquality;

            stopit = result_insert(result);
          }

        result_destroy(tmpresult);
        free2(newdata);
      }
    }

  result_clean(result);
  if(result_depth(result)==0)
    RAISE(fatal_XCEPT, NULL, NULL);
  return result;
}

/*==========================================================================
//  Destruction des fichiers de donnees frequencielles
//========================================================================*/

void
auto_remove(void)
{
  shift_remove();
  return;
}

/*========================================================================*/
