/*==========================================================================
//  Dcryptage du chiffrement de Vigenre
//  Projet TER, Matrise d'Informatique 1997,
//  Universit de Bordeaux 1, France,
//  Tous droits rservs.
//
//  de Maillard Luc, Montiton Frdric
//  Morreeuw Jean-Franois et Lompo Issa
//
//  mailto:morreeuw@emi.u-bordeaux.fr,dbv@komicom.remcomp.fr
//  http://www.emi.u_bordeaux.fr/~morreeuw/universite/TER
//
//  Module de skiplist permettant un accs par chanage ou par cl. Ce
//  module est une version lgrement modifie du module ralis pour le
//  projet de licence LMI en 1996.
//
//========================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include "dsl.h"

/*========================================================================*/

typedef struct dsl_cell *dsl_cell;

struct dsl_cell
{
  void *key;
  void *value;
  dsl_cell *forward;
};

/*------------------------------------------------------------------------*/

struct dsl
{
  int level;
  dsl_cell header;
  int (*comparefun)(void *key1, void *key2);
};

/*------------------------------------------------------------------------*/

static struct dsl_cell dsl_NIL = { NULL, NULL, NULL};

/*========================================================================*/

static int       compare(int (*comparefun)(void *, void *),
			 void *key1, void *key2);

static int       _level();
static dsl_cell *_expand(dsl_cell array[], int level);
static dsl_cell  _cell(int level, void *key, void *value);
static dsl_cell  _find(dsl head, void *key);

/*========================================================================*/

int
compare(int (*comparefun)(void *, void *), void *key1, void *key2)
{
  if(!key1)
    return 0;
  if(!key2)
    return 1;
  return comparefun(key1, key2);
}

/*========================================================================*/

dsl
dsl_create(int (*comparefun)(void *, void *))
{
  dsl head = (dsl) malloc(sizeof *head);

  head->level = 0;
  head->comparefun = comparefun;
  head->header = (dsl_cell) malloc(sizeof *head->header);
  head->header->key = NULL;
  head->header->value = NULL;
  head->header->forward = (dsl_cell *) malloc(sizeof *head->header->forward);
  head->header->forward[0] = &dsl_NIL;
  return head;
}

/*========================================================================*/

void
dsl_destroy(dsl head)
{
  free(head->header->forward);
  free(head->header);
  free(head);
  return;
}

/*========================================================================*/

int
dsl_member(dsl head, void *key)
{
  return _find(head, key)?1:0;
}

/*------------------------------------------------------------------------*/

void *
dsl_find(dsl head, void *key)
{
  dsl_cell cell = _find(head, key);
  return cell?cell->value:NULL;
}

/*------------------------------------------------------------------------*/

void *
dsl_next(dsl head, void *key)
{
  dsl_cell cell;
  if(key)
    cell = _find(head, key);
  else
    cell = head->header;

  if(cell)
    cell = cell->forward[0];

  if(cell)
    return cell->key;
  else
    return NULL;
}

/*------------------------------------------------------------------------*/

dsl_cell
_find(dsl head, void *key)
{
  int i;
  dsl_cell cell = head->header;

  for( i=head->level; i>=0; i--)
    while( compare(head->comparefun, cell->forward[i]->key, key ) )
      cell = cell->forward[i];
  cell = cell->forward[0];

  return compare( head->comparefun, key, cell->key )?NULL:cell;
}

/*========================================================================*/

void
dsl_insert(dsl head, void *key, void *value)
{
  int i;
  int level = _level();
  dsl_cell cell = head->header;
  dsl_cell *update = (dsl_cell *) malloc( (head->level+2) * sizeof *update );
  
  for( i=head->level; i>=0; update[i--]=cell)
    while( compare(head->comparefun, cell->forward[i]->key, key ) )
      cell = cell->forward[i];

  if( level > head->level )
    {
      level = ++head->level;
      update[level] = head->header;
      head->header->forward = _expand( head->header->forward, level );
    }
  cell = _cell( level, key, value);
  
  for(i=0; i<=level; i++)
    {
      cell->forward[i] = update[i]->forward[i];
      update[i]->forward[i] = cell;
    }

  free(update);
  return;
}

/*------------------------------------------------------------------------*/

#define randomP 16384

int
_level()
{
  int level = 0;

  while(rand() < randomP)
    level++;

  return level;
}

/*------------------------------------------------------------------------*/

dsl_cell *
_expand(dsl_cell array[], int level)
{
  int i;
  dsl_cell *newarray = (dsl_cell *) malloc( (level+1) * sizeof **newarray );

  for(i=0; i<level; i++)
    newarray[i] = array[i];

  newarray[ level ] = &dsl_NIL;

  free(array);
  return newarray;
}

/*------------------------------------------------------------------------*/

dsl_cell
_cell(int level, void *key, void *value)
{
  dsl_cell cell = (dsl_cell) malloc(sizeof *cell);
 
  cell->key = key;
  cell->value = value;
  cell->forward = (dsl_cell *) malloc( (level+1) * sizeof *cell->forward );
  return cell;
}

/*========================================================================*/

void
dsl_delete(dsl head, void *key)
{
  int i;
  dsl_cell cell = head->header;
  dsl_cell *update = (dsl_cell *) malloc( (head->level+1) * sizeof *update );
  
  for(i=head->level; i>=0; update[i--]=cell)
    while( compare( head->comparefun, cell->forward[i]->key, key ) )
      cell = cell->forward[i];
  cell = cell->forward[0];

  for(i=0; (i <= head->level) && (update[i]->forward[i] == cell); i++)
    update[i]->forward[i] = cell->forward[i];
  
  while( head->level && !head->header->forward[head->level]->forward )
    head->level--;
  
  free(cell->forward);
  free(cell);
  free(update);
  return;
}

/*========================================================================*/
