/*==========================================================================
//  Dcryptage du chiffrement de Vigenre
//  Projet TER, Matrise d'Informatique 1997,
//  Universit de Bordeaux 1, France,
//  Tous droits rservs.
//
//  de Maillard Luc, Montiton Frdric
//  Morreeuw Jean-Franois et Lompo Issa
//
//  mailto:morreeuw@emi.u-bordeaux.fr,dbv@komicom.remcomp.fr
//  http://www.emi.u_bordeaux.fr/~morreeuw/universite/TER
//
//  MODULE DE RECHERCHE DES DECALAGES RELATIFS DES LETTRES DE LA CLE
//
//========================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include "kh.h"

#define KH_WIDTH_MAX 4

typedef struct equation *equation_p;
typedef struct linear *linear_p;

struct equation
{
  double *quality;
  short * order;
  short   position;
  short   shift;
};

struct linear
{
  short       width;
  short *     choice;
  equation_p *table;
  result_p    result;
};

static result_p _analyse(char *data, long length, short cycle,
			 short start, short distance, int depth);
static void     _frequency_fill(linear_p linear, char *data,
				long length, short cycle, short start);
static void     _equation_order(linear_p linear);
static void     _find_solutions(linear_p linear, short *choice, short width);
static int      _next_choice(short *selection, short number, short width);
static int      _test_solution(linear_p linear);
static void     _choice_project(linear_p linear);
static void     _choice_spread(linear_p linear);
static double   _choice_quality(linear_p linear);
static double   _choice_maximum(linear_p linear);
static void     _filter_solutions(result_p result);

/*==========================================================================
//  Fonction d'appel externe, la fonction principale tant rcursive
//========================================================================*/

result_p
kh_compute(char *data, long length, short cycle, long depth)
{
  result_p result;

  VERIFY(data);

  result = _analyse(data, length, cycle, 0, cycle+1, depth);
  _filter_solutions(result);
  return result;
}

/*==========================================================================
//  Noyau de la recherche des dcalages. La mthode consiste soit a
//  scinder la recherche et  effectuer la jointure, soit  calculer
//  directement les solutions les plus probables du systme.
//========================================================================*/

result_p
_analyse(char *data, long length, short cycle, short start,
	 short distance, int depth)
{
  result_p result;

  if(distance <= KH_WIDTH_MAX)
    {
      int i;
      struct linear _linear;
      linear_p linear = &_linear;
      equation_p table[(KH_WIDTH_MAX*(KH_WIDTH_MAX+1))/2];
      struct equation table_map[(KH_WIDTH_MAX*(KH_WIDTH_MAX+1))/2];
      double quality_map[CHAR_RANGE*(KH_WIDTH_MAX*(KH_WIDTH_MAX+1))/2];
      short order_map[CHAR_RANGE*(KH_WIDTH_MAX*(KH_WIDTH_MAX+1))/2];
      short choice[KH_WIDTH_MAX-1];

      distance = KH_WIDTH_MAX;

      linear->width = distance-1;

      for(i=0; i<distance-1; i++)
        choice[i]=0;
      linear->choice = choice;

      for(i=0; i<(KH_WIDTH_MAX*(KH_WIDTH_MAX+1))/2; i++)
        {
          table[i] = &table_map[i];
          table[i]->quality = &quality_map[CHAR_RANGE*i];
          table[i]->order = &order_map[CHAR_RANGE*i];
          table[i]->position = 0;
          table[i]->shift = 0;
        }
      linear->table = table;

      linear->result = NULL;

      _frequency_fill(linear, data, length, cycle, start);
      _equation_order(linear);
      result = (linear->result = result_create(depth, distance-1));

      {
        short newchoice[KH_WIDTH_MAX-1];
        for(i=0; i<KH_WIDTH_MAX-1; i++)
          newchoice[i]=linear->choice[i]+(i*(i+1))/2;
        _find_solutions(linear, newchoice, 0);
      }
    }
  else
    {
      result_p result1;
      result_p result2;
      long width1 = distance/2;
      long width2 = distance-width1;
      long distance1 = width1+1;
      long distance2 = width2+1;

      result1 = _analyse(data, length, cycle, start, distance1, depth);
      result2 = _analyse(data, length, cycle,
			 start+width1-1, distance2, depth);

      result = result_join(result1, width1, 0, width1-1,
			   result2, width2-1, 1, 0);

      result_destroy(result1);
      result_destroy(result2);
    }

  if(result_clean(result)==0)
    RAISE(fatal_XCEPT, NULL, NULL);
  result_normalize(result);

  return result;
}

/*==========================================================================
//  Initialisation du systme : cration des tables de coefficients
//  Ces coefficients correspondent aux qualits des quations dcrivant
//  les dcalages relatifs des lettres de la cl
//========================================================================*/

void
_frequency_fill(linear_p linear, char *data, long length,
		short cycle, short start)
{
  short i, j, k;
  short distance = linear->width+1;
  double table[KH_WIDTH_MAX][CHAR_RANGE];

  for(i=0; i<distance; i++)
    ic_compute(table[i], 1, data, length, cycle, (start+i)%cycle);

  for(i=1; i<distance; i++)
    for(j=i; j>0; j--)
      for(k=0; k<CHAR_RANGE; k++)
        linear->table[(i*(i+1))/2-j]->quality[k]
	  = ic_mutual(table[j-1], table[i], 1, k);

  return;
}

/*==========================================================================
//  Initialisation du systme : calcul des tables ordonnes de dcalage
//  Les quations de dcalages relatifs sont tries suivant les qualits
//========================================================================*/

static int _compare(const void *key1, const void *key2);

/*------------------------------------------------------------------------*/

int
_compare(const void *key1, const void *key2)
{
  double value1 = *(((double *) key1)+1);
  double value2 = *(((double *) key2)+1);

  if(value1>value2)
    return -1;
  else if(value1<value2)
    return 1;
  else
    return 0;
}

/*------------------------------------------------------------------------*/

void
_equation_order(linear_p linear)
{
  int i, j;
  short width = linear->width;
  int height = (width*(width+1))/2;
  double table[CHAR_RANGE*2];

  for(i=0; i<height; i++)
    {
      for(j=0; j<CHAR_RANGE; j++)
        {
          table[2*j]=(double)j;
          table[2*j+1]=linear->table[i]->quality[j];
        }

      qsort((void *) table, CHAR_RANGE, 2*sizeof(*table), _compare);

      for(j=0; j<CHAR_RANGE; j++)
        linear->table[i]->order[j] = (short) table[2*j];
    }

  return;
}

/*==========================================================================
//  Recherche probabiliste des meilleures combinaisons. La limitation des
//  recherches est realise par l'appel  _test_solution() qui retourne -1
//  ds qu'il n'y a plus moyen de trouver un meilleur rsultat.
//========================================================================*/

void
_find_solutions(linear_p linear, short *choice, short width)
{
  int stop=0;
  int step=-1;
  short i;
  short selection[KH_WIDTH_MAX-1];
  short newchoice[KH_WIDTH_MAX-1];
  
  for(stop=0, step=0; stop!=-1; step++)
    {
      if(width==0)
        {
          width = linear->width;
          step=-1;
        }
      else
        for(i=0; i<width; i++)
          if( (++(linear->table[choice[i]]->position)) >= CHAR_RANGE)
            stop=-1;

      if(stop!=-1)
        stop = _test_solution(linear);

      if(stop!=-1)
        {
          for(i=0; i<width; i++)
            selection[i]=i;

          for(i=1; i<width; i++)
            {
              int next=0;
              for(; next!=-1; next=_next_choice(selection, width, i))
                {
                  long j;
                  for(j=0; j<i; j++)
                    newchoice[j]=choice[selection[j]];
                  _find_solutions(linear, newchoice, i);
                }
            }
        }
    }

  for(i=0; i<width; i++)
    (linear->table[choice[i]]->position)-=step;

  return;
}

/*--------------------------------------------------------------------------
//  Choix de width quations parmi la slection de number quations. Ce
//  choix est ralise directement dans le tableau le dcrivant...
//------------------------------------------------------------------------*/

int
_next_choice(short *selection, short number, short width)
{
  short last = width-1;

  if(width!=1 && _next_choice(selection, selection[last], width-1)!=-1)
    return 0;
  else
    if( ++selection[last] < number)
      return 0;
    else
      {
        selection[last]=last;
        return -1;
      }
}

/*==========================================================================
//  Insertion dans la table si la probabilit est bonne et teste le maximum
//  Dans le cas ou la majoration indique que la solution ne saurait en
//  aucun cas entrer dans le resultat final, retourne -1
//========================================================================*/

int
_test_solution(linear_p linear)
{
  short i;
  double quality, maximum;
  short *newoffset = result_newoffset(linear->result);
  double *newquality = result_newquality(linear->result);

  _choice_project(linear);
  _choice_spread(linear);
  quality = _choice_quality(linear);
  maximum = _choice_maximum(linear);

  if(*result_quality(linear->result, result_depth(linear->result)-1) > maximum)
    return -1;

  for(i=0; i<linear->width; i++)
    newoffset[i] = linear->table[(i*(i+1))/2]->shift;
  newquality[0] = quality;

  result_insert(linear->result);
  return 0;
}

/*--------------------------------------------------------------------------
//  Met  jour les dcalages pour les quations tudies. Typiquement cette
//  fonction ne sert  rien lorsque les quations principales sont les
//  quations dcalages entre lettres conscutives, mais comme rien dans
//  le module n'interdit que ce ne soit plus le cas un jour...
//------------------------------------------------------------------------*/

void
_choice_project(linear_p linear)
{
  short i;

  for(i=0; i < linear->width; i++)
    {
      long j = (i*(i+1))/2+linear->choice[i];
      linear->table[j]->shift
	= linear->table[j]->order[ linear->table[j]->position ];
    }

  return;
}

/*--------------------------------------------------------------------------
//  Rpercute les dcalages des quations principales sur l'ensemble
//  des quations du systme. A utiliser avant _choice_quality()
//------------------------------------------------------------------------*/

void
_choice_spread(linear_p linear)
{
  short i;
  short *choice = linear->choice;

  for(i=0; i < linear->width; i++)
    {
      short j;
      int index = (i*(i+1))/2;
 
      linear->table[index]->shift = linear->table[index+choice[i]]->shift;

      for(j=1; j<=choice[i]; j++)
        linear->table[index]->shift -= linear->table[((i-j)*(i-j+1))/2]->shift;

      if(linear->table[index]->shift < 0)
        linear->table[index]->shift
	  += CHAR_RANGE*(1-(linear->table[index]->shift)/CHAR_RANGE);

      for(j=0; j<i; j++)
        linear->table[index+j+1]->shift
	  = (linear->table[index]->shift + linear->table[index-i+j]->shift)
	  %CHAR_RANGE;
    }

  return;
}

/*--------------------------------------------------------------------------
//  Calcul du coefficient global pour l'ensemble des dcalages choisis.
//  Ce coefficient caractrise une solution particulire dans le systme.
//------------------------------------------------------------------------*/

double
_choice_quality(linear_p linear)
{
  int i;
  short width = linear->width;
  int height = (width*(width+1))/2;
  double result=1;

  for(i=0; i<height; i++)
    result *= linear->table[i]->quality[linear->table[i]->shift];

  return result;
}

/*--------------------------------------------------------------------------
//  Calcul d'un maximum local pour limiter la recherche. Ce coefficient
//  est une majoration du coefficient global en n'utilisant que les
//  quations principales.
//------------------------------------------------------------------------*/

double
_choice_maximum(linear_p linear)
{
  short i;
  double result=1;

  for(i=0; i<linear->width; i++)
    {
      short j;
      int index = (i*(i+1))/2;

      for(j=0; j<=i; j++)
        if(linear->choice[i]!=j)
          result *= linear->table[index+j]->
	    quality[linear->table[index+j]->order[0]];
        else
          result
	    *= linear->table[index+j]->quality[linear->table[index+j]->shift];
    }

  return result;
}

/*==========================================================================
//  Filtre les rsultats non valides. La premire tape consiste  trouver
//  le minimum des rsultats compatibles, la seconde  multiplier les
//  rsultats non compatibles par ce coefficient pour reduire l'importance
//========================================================================*/

static double _mapminimum(short *offset, double quality,
			  short width, void *argument);

double
_mapminimum(short *offset, double quality, short width, void *argument)
{
  short i;
  long value = 0;

  for(i=0; i<width; i++)
    value += offset[i];
  value %= CHAR_RANGE;

  if(value==0 && *((double *) argument)>quality)
    *((double *) argument) = quality;

  return quality;
}

/*------------------------------------------------------------------------*/

static double _mapreduce(short *offset, double quality,
			 short width, void *argument);

double
_mapreduce(short *offset, double quality, short width, void *argument)
{
  short i;
  long value = 0;

  for(i=0; i<width; i++)
    value += offset[i];
  value %= CHAR_RANGE;

  if(value==0)
    return quality;
  else
    return quality*(*((double *) argument));
}

/*------------------------------------------------------------------------*/

void
_filter_solutions(result_p result)
{
  double minimum = 1;

  result_map(result, _mapminimum, (void *) &minimum);
  if(minimum==0)
    minimum=1;
  result_map(result, _mapreduce, (void *) &minimum);
  result_sort(result);
  result_clean(result);

  return;
}

/*========================================================================*/
