/*==========================================================================
//  Dcryptage du chiffrement de Vigenre
//  Projet TER, Matrise d'Informatique 1997,
//  Universit de Bordeaux 1, France,
//  Tous droits rservs.
//
//  de Maillard Luc, Montiton Frdric
//  Morreeuw Jean-Franois et Lompo Issa
//
//  mailto:morreeuw@emi.u-bordeaux.fr,dbv@komicom.remcomp.fr
//  http://www.emi.u_bordeaux.fr/~morreeuw/universite/TER
//
//  MODULE IMPLANTANT UNE STRUCTURE COMMUNE POUR LES RESULTATS
//
//========================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include "result.h"

struct result
{
  int depth;
  short width;
  short **offset;
  double *quality;

  short *offset_map;
};

static int _compare(short *offset1, short *offset2, short width);
static int _store(result_p result, short *offset1, short width1, short
		  *offset2, short width2, double quality);

/*==========================================================================
//  Fabrication de la table des rsultats.
//========================================================================*/

result_p
result_create(int depth, short width)
{
  int i;
  result_p result = (result_p) malloc2(sizeof(*result));
  result->depth = depth;
  result->width = width;
  result->offset = (short **) malloc2((depth+1)*sizeof(*result->offset));
  result->offset_map
    = (short *) malloc2((depth+1)*width*(sizeof(*result->offset_map)));
  result->quality = (double *) malloc2((depth+1)*sizeof(*result->quality));
  for(i=0; i<depth+1; i++)
    {
      result->offset[i]= &result->offset_map[i*width];
      result->quality[i]=0;
    }

  return result;
}

/*==========================================================================
//  Destruction de la table des rsultats.
//========================================================================*/

void
result_destroy(result_p result)
{
  VERIFY(result);

  free2(result->offset);
  free2(result->offset_map);
  free2(result->quality);
  free2(result);

  return;
}

/*==========================================================================
//  Insertion dans la table des rsultats. La table de rsultat ne procde
//   aucune allocation => pour les algorithmes critiques. Le principe
//  repose sur une zone rserve dont l'utilisateur demande l'adresse. Si
//  la zone doit tre insre, la zone sortant est positionne comme
//  nouvelle zone reserve. Toute insertion est ordonne, et par consquent
//  la table doit tre au pralable dj ordonne.
//========================================================================*/

int
_compare(short *offset1, short *offset2, short width)
{
  short i;

  VERIFY(offset1);
  VERIFY(offset2);

  for(i=0; i<width; i++)
    if(offset1[i]!=offset2[i])
      return 0;
  return 1;
}

/*------------------------------------------------------------------------*/

int
result_insert(result_p result)
{
  int i, j, k;
  short width;
  int depth;
  short *offset_tmp;
  double quality_tmp;
  short *offset_new;
  double quality_new;

  VERIFY(result);

  width = result->width;
  depth = result->depth;
  offset_new = result->offset[depth];
  quality_new = result->quality[depth];

  if(result->quality[depth-1] >= quality_new)
    return -1;

  for(i=0; result->quality[i] >= quality_new; i++)
    if(_compare(result->offset[i], offset_new, width)==1)
      return 0;

  for(j=i, k=i; j<depth; j++)
    if(j==i
       || result->quality[j]==0
       || _compare(result->offset[j], result->offset[i], width)!=1)
      {
        offset_tmp = result->offset[j];
        result->offset[k] = offset_new;
        offset_new = offset_tmp;
        quality_tmp = result->quality[j];
        result->quality[k] = quality_new;
        quality_new = quality_tmp;
        k++;
      }
    else
      result->offset[depth] = result->offset[k];

  if(k==j)
    result->offset[depth] = offset_new;
  else
    result->depth -= j-k+1;

  return 0;
}

/*==========================================================================
//  Remise  1 du maximum de la table des rsultats.
//========================================================================*/

void
result_normalize(result_p result)
{
  double maximum = 0;
  int i;

  VERIFY(result);

  for(i=0; i<result->depth; i++)
    if(result->quality[i] > maximum)
      maximum = result->quality[i];

  for(i=0; i<result->depth; i++)
    result->quality[i] /= maximum;

  return;
}

/*==========================================================================
//  Elimine les entres vides.
//========================================================================*/

int
result_clean(result_p result)
{
  VERIFY(result);

  for(; result->depth>0 && result->quality[result->depth-1]==0; --result->depth)
    result->offset[result->depth-1] = NULL;

  return result->depth;
}

/*==========================================================================
//  Applique une fonction au contenu de la table des rsultats.
//========================================================================*/

void
result_map(result_p result, double (*function)(short *offset, double quality,
					       short width, void
					       *argument),
	   void *argument)
{
  short i;

  VERIFY(result);

  for(i=0; i<result->depth && result->quality[i]!=0; i++)
    result->quality[i] = function(result->offset[i],
				  result->quality[i], result->width, argument);
  return;
}

/*==========================================================================
//  Trie le contenu de la table. Typiquement, cette fonction est appele
//  aprs l'application d'une fonction par result_map().
//========================================================================*/

void
result_sort(result_p result)
{
  int i, j;

  VERIFY(result);

  for(i=0; i<result->depth-1; i++)
    for(j=i+1; j<result->depth; j++)
      if(result->quality[i] < result->quality[j])
        {
          short *offset_tmp = result->offset[i];
          double quality_tmp = result->quality[i];
          result->offset[i] = result->offset[j];
          result->quality[i] = result->quality[j];
          result->offset[j] = offset_tmp;
          result->quality[j] = quality_tmp;
        }

  return;
}

/*==========================================================================
//  Rcupration de la zone rserve pour le tampon d'insertion
//========================================================================*/

short *
result_newoffset(result_p result)
{
  short i;
  VERIFY(result);
  for(i=0; i<result->width; i++)
    result->offset[result->depth][i]=0;
  return result->offset[result->depth];
}

/*------------------------------------------------------------------------*/

double *
result_newquality(result_p result)
{
  VERIFY(result);
  result->quality[result->depth]=0;
  return &result->quality[result->depth];
}

/*==========================================================================
//  Lecture du contenu de la table
//========================================================================*/

int
result_depth(result_p result)
{
  VERIFY(result);
  return result->depth;
}

/*------------------------------------------------------------------------*/

short *
result_offset(result_p result, int index)
{
  VERIFY(result);
  return result->offset[index];
}

/*--------------------------------------------------------------------------
//  Cette fonction permet de faire un effet de bord. A utiliser avec soin
//  car la table doit etre ordonne pour l'insertion/jointure.
//------------------------------------------------------------------------*/

double *
result_quality(result_p result, int index)
{
  VERIFY(result);
  return &result->quality[index];
}

/*==========================================================================
//  Effectue la jointure de deux tables de rsultats
//========================================================================*/

result_p
result_join(result_p result1, short width1, short start1, short join1,
	    result_p result2, short width2, short start2, short join2)
{
  long i, j;
  result_p result;
  double minimum = 1;
  short stopit;

  VERIFY(result1);
  VERIFY(result2);

  result = result_create((result1->depth<result2->depth)
			 ?result1->depth:result2->depth, width1+width2);

  for(i=0; i<result1->depth; i++)
    for(stopit=0, j=0; j<result2->depth && stopit!=-1; j++)
      if(result1->offset[i][join1]==result2->offset[j][join2])
        {
          double quality = result1->quality[i]*result2->quality[j];
          if(minimum>quality)
            minimum = quality;
          stopit = _store(result, result1->offset[i]+start1, width1,
			  result2->offset[j]+start2, width2, quality);
        }

  if(result->quality[result->depth-1]!=0)
    return result;

  for(i=0; i<result1->depth; i++)
    for(stopit=0, j=0; j<result2->depth && stopit!=-1; j++)
      if(result1->offset[i][join1]!=result2->offset[j][join2])
        stopit = _store(result, result1->offset[i]+start1, width1,
			result2->offset[j]+start2, width2,
			(result1->quality[i]*result2->quality[j])*minimum);
  return result;
}

/*--------------------------------------------------------------------------
//  Enregistre la jointure de deux enregistrements
//------------------------------------------------------------------------*/

int
_store(result_p result, short *offset1, short width1,
       short *offset2, short width2, double quality)
{
  short i;
  short *newoffset = result_newoffset(result);
  double *newquality = result_newquality(result);

  for(i=0; i<width1+width2; i++)
    newoffset[i] = (i<width1)?offset1[i]:offset2[i-width1];
  newquality[0] = quality;

  return result_insert(result);
}

/*========================================================================*/

void
result_view(result_p result)
{
  int i;

  VERIFY(result);

  for(i=0; i<result->depth; i++)
    {
      short j;
      if(result->offset[i])
        for(j=0; j<result->width; j++)
          printf("%.2d ", result->offset[i][j]);
      printf(" %g\n", result->quality[i]);
    }

  return;
}

/*========================================================================*/
