/*==========================================================================
//  Dcryptage du chiffrement de Vigenre
//  Projet TER, Matrise d'Informatique 1997,
//  Universit de Bordeaux 1, France,
//  Tous droits rservs.
//
//  de Maillard Luc, Montiton Frdric
//  Morreeuw Jean-Franois et Lompo Issa
//
//  mailto:morreeuw@emi.u-bordeaux.fr,dbv@komicom.remcomp.fr
//  http://www.emi.u_bordeaux.fr/~morreeuw/universite/TER
//
//  MODULE DE DECRYPTAGE DU CHIFFREMENT DE CAESAR (MONOALPHABETIQUE)
//
//========================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>

#if !(__dest_os) || __dest_os != __be_os
#include <unistd.h>
#endif

#include "shift.h"

typedef struct language *lgg_p;

struct language
{
  char *name;
  double uletter[CHAR_RANGE];
  double dletter[CHAR_RANGE*CHAR_RANGE];
};

static lgg_p reference[LANGUAGE_MAX];

static void   _load(char *filename);
static double _sqr(double value);
static void   _binary2double(unsigned char *binary_table,
			     double *double_table);

/*========================================================================*/

short
shift_install(char *path)
{
  short i;
  int position;
  struct stat buffer;
  struct dirent *entry;
  DIR *dirp = opendir(path);
  char *pathname;

  if(dirp==NULL)
    RAISE(io_XCEPT, path, NULL);

  pathname = (char *) malloc2(strlen(path)+1+MAXNAMLEN);

  strcpy(pathname, path);
  if(pathname[(position=strlen(path))-1]!='/')
    pathname[position++]='/';

  for(i=0; i<LANGUAGE_MAX; i++)
    reference[i]=NULL;

  while( (entry=readdir(dirp)) )
    {
      strcpy(pathname+position, entry->d_name);
      if(stat(pathname, &buffer)==-1)
        RAISE(io_XCEPT, pathname, NULL);
      if((buffer.st_mode & S_IFMT)==S_IFREG
	 && buffer.st_size == CHAR_RANGE*CHAR_RANGE*LANGUAGE_DATA_BYTES)
        _load(pathname);
    }

  if(closedir(dirp)==-1)
    RAISE(io_XCEPT, path, NULL);
  free2(pathname);

  for(i=0; i<LANGUAGE_MAX && reference[i]!=NULL; i++)
    ;
  return i;
}

/*------------------------------------------------------------------------*/

void
_load(char *filename)
{
  short i, j;
  int fd;
  short position;

  VERIFY(filename);

  for(position=0; reference[position]!=NULL; position++)
    if(position==LANGUAGE_MAX-1)
      return;

  reference[position] = (lgg_p) malloc2(sizeof(*reference[position]));

  reference[position]->name = (char *) strdup2(filename);

  {
    unsigned char binary[CHAR_RANGE*CHAR_RANGE*LANGUAGE_DATA_BYTES];

    if((fd=open(filename, O_RDONLY))==-1 ||
       read(fd, binary, CHAR_RANGE*CHAR_RANGE*LANGUAGE_DATA_BYTES)
       != CHAR_RANGE*CHAR_RANGE*LANGUAGE_DATA_BYTES || close(fd)==-1)
      RAISE(io_XCEPT, filename, NULL);

    _binary2double(binary, reference[position]->dletter);
  }

  for(i=0; i<CHAR_RANGE; i++)
    for(j=0, reference[position]->uletter[i]=0; j<CHAR_RANGE; j++)
      reference[position]->uletter[i]
	+= reference[position]->dletter[i*CHAR_RANGE+j];

  return;
}

/*==========================================================================
//  Recherche le dcalage d'un chiffrement monoalphabtique. Les langues
//  doivent avoir t initialement lues.
//========================================================================*/

double
_sqr(double value)
{
  return value*value;
}

/*------------------------------------------------------------------------*/

result_p
shift_search(char *data, long length, int depth)
{
  short position;
  struct language language = {NULL};
  result_p result = result_create(depth, 2);

  VERIFY(data);
  VERIFY(reference[0]);

  ic_compute(language.uletter, 1, data, length, 1, 0);
  ic_compute(language.dletter, 2, data, length, 1, 0);

  for(position=0; position<LANGUAGE_MAX
	&& reference[position]!=NULL; position++)
    {
      short i;

      for(i=0; i<CHAR_RANGE; i++)
        {
          double uquality = ic_mutual(reference[position]->uletter,
				      language.uletter, 1, i);
          double dquality = ic_mutual(reference[position]->dletter,
				      language.dletter, 2, i);
          double quality = _sqr(uquality*_sqr(dquality));
          short *newoffset = result_newoffset(result);
          double *newquality = result_newquality(result);

          newoffset[0] = i;
          newoffset[1] = position;
          newquality[0] = quality;
          result_insert(result);
        }
    }

  return result;
}

/*========================================================================*/

char *
shift_language(short index)
{
  if(index>LANGUAGE_MAX || reference[index]==NULL)
    RAISE(fatal_XCEPT, "reference", "Data access exception");
  return reference[index]->name;
}

/*========================================================================*/

void
shift_remove(void)
{
  short position;

  for(position=0; reference[position]!=NULL; reference[position++]=NULL)
    {
      free2(reference[position]->name);
      free2(reference[position]);
    }

  return;
}

/*==========================================================================
//  Les deux fonctions convertissent des donnes entre format binaire et
//  rel (double). Ce travail est ncessaire pour lire les fichiers binaires
//  contenant les frquences des langues pour assurer la portabilit sur
//  les machines non normalises (pc pentium par exemple). Le format
//  utilis est celui des flottants simple prcision de la norme IEEE 754.
//  (D.Patterson/J.Hennessy, "Organisation et conception des ordinateurs",
//  Dunod 1994, page 234).
//========================================================================*/

void
shift_double2binary(double *double_table, unsigned char *binary_table)
{
  short i;

  for(i=0; i<CHAR_RANGE*CHAR_RANGE; i++)
    {
      double value = double_table[i];
      unsigned char double_s = 0;
      unsigned char double_e = 0;
      long double_m = 0;

      if(value<0)
        {
          double_s = 1;
          value *= -1;
        }

      if(value>=2)
        for(double_e=127; value>=2 && double_e!=0xff; double_e++)
          value /= 2;
      else if(value<1)
        for(double_e=127; value<1 && double_e!=0x00; double_e--)
          value *= 2;
      if(value>=1 && value<2)
        value -= 1;
      else
        value=0;

      double_m = value*pow(2, 23);

      binary_table[LANGUAGE_DATA_BYTES*i] = (double_s<<7) | (double_e>>1);
      binary_table[LANGUAGE_DATA_BYTES*i+1]
	= (double_e<<7) | ((double_m>>16)&0x7f);
      binary_table[LANGUAGE_DATA_BYTES*i+2] = (double_m>>8)&0xff;
      binary_table[LANGUAGE_DATA_BYTES*i+3] = double_m&0xff;
    }

  return;
}

/*------------------------------------------------------------------------*/

void
_binary2double(unsigned char *binary_table, double *double_table)
{
  short i;

  for(i=0; i<CHAR_RANGE*CHAR_RANGE; i++)
    {
      double value = 0;
      unsigned char double_s = binary_table[LANGUAGE_DATA_BYTES*i]>>7;
      unsigned char double_e = (binary_table[LANGUAGE_DATA_BYTES*i]<<1)
                               | (binary_table[LANGUAGE_DATA_BYTES*i+1]>>7);
      long double_m = (((long) binary_table[LANGUAGE_DATA_BYTES*i+1])&0x7f)<<16
                      | ((long) binary_table[LANGUAGE_DATA_BYTES*i+2])<<8
                      | ((long) binary_table[LANGUAGE_DATA_BYTES*i+3]);
      value = pow(-1, (double) double_s)
	      *(1+((double) double_m)*pow(2, -23))
	      *pow(2, (double) (((long) double_e)-127));

      if(double_e==0x00)
        value = 0;
      double_table[i] = value;
    }

  return;
}

/*========================================================================*/
